from torch.utils.data import Dataset


class GenericGenerator(Dataset):
    def __init__(self, dataset):
        """
        A generic data generator which can be used to build preprocessing and data augmentation pipelines.
        The data generator subclasses the pytorch Dataset class and can therefore be used directly with DataLoaders
        in pytorch. The processing pipeline of the generator is defined through a series of processing steps or
        augmentations. For each data sample, the generator calls the augmentations in order.
        Information between the augmentation steps is passed through a state dict.
        The state dict is a python dictionary mapping keys to a tuple (data, metadata).
        In getitem, the generator automatically populates the initial dictionary with the waveforms
        and the corresponding metadata for the row from the underlying data set using the key "X".
        After applying all augmentation, the generator removes all metadata information.
        This means that the output dict only maps keys to the data part.
        Any metadata that should be output needs to explicitly be written to data.

        Augmentation can be either callable classes of functions.
        Functions are usually best suited for simple operations, while callable classes offer more configuration
        options. SeisBench already offers a set of standard augmentations for augmentation and preprocessing,
        e.g., for window selection, data normalization or different label encodings,
        which should cover many common use cases.
        For details on implementing custom augmentations we suggest looking at the examples provided.

        SeisBench augmentations by default always work on the key "X".
        Label generating augmentations by default put labels into the key "y".
        However, for more complex workflows, the augmentations can be adjusted using the key argument.
        This allows in particular none-sequential augmentation sequences.

        :param dataset: The underlying SeisBench data set.
        :type dataset: seisbench.data.WaveformDataset or seisbench.data.MultiWaveformDataset
        """
        self._augmentations = []
        self.dataset = dataset
        super().__init__()

    def augmentation(self, f):
        """
        Decorator for augmentations.
        """
        self._augmentations.append(f)

        return f

    def add_augmentations(self, augmentations):
        """
        Adds a list of augmentations to the generator. Can not be used as decorator.

        :param augmentations: List of augmentations
        :type augmentations: list[callable]
        """
        if not isinstance(augmentations, list):
            raise TypeError(
                "The argument of add_augmentations must be a list of augmentations."
            )

        self._augmentations.extend(augmentations)

    def __str__(self):
        summary = f"GenericGenerator with {len(self._augmentations)} augmentations:\n"
        for i, aug in enumerate(self._augmentations):
            summary += f" {i + 1}.\t{str(aug)}\n"
        return summary

    def __len__(self):
        return len(self.dataset)

    def __iter__(self):
        return self

    def __getitem__(self, idx):
        state_dict = {"X": self.dataset.get_sample(idx)}

        # Recursive application of augmentation processing methods
        for func in self._augmentations:
            func(state_dict)

        # Remove all metadata from the output
        state_dict = {k: v[0] for k, v in state_dict.items()}

        return state_dict


class SteeredGenerator(GenericGenerator):
    """
    This data generator follows the same principles as the :py:func:`~GenericGenerator`.
    However, in contrast to the :py:func:`~GenericGenerator` the generator is controlled by a dataframe with control
    information. Each row in the control dataframe corresponds to one example output by the generator.
    The dataframe holds two types of information. First, information identifying the traces, provided using the
    `trace_name` (required), `trace_chunk` (optional), and `trace_dataset` (optional). See the description of
    :py:func:`~seisbench.data.base.WaveformDataset.get_idx_from_trace_name` for details.
    Second, additional information for the augmentations.
    This additional information is stored in `state_dict["_control_"]` as a dict.
    This generator is particularly useful for evaluation, e.g., when extracting predefined windows from a trace.

    Note that the "_control_" group will usually not be modified by augmentations.
    This means, that for example after a window selection, sample positions might be off.
    To automatically handle these, you must explicitly put the relevant control information
    into the state_dict metadata of the relevant key.

    .. warning::
        This generator should in most cases not be used for changing label distributions by resampling the dataset.
        For this application, we recommend using a
        `pytorch Sampler <https://pytorch.org/docs/stable/data.html#data-loading-order-and-sampler>`_.

    :param dataset: The underlying SeisBench data set
    :type dataset: seisbench.data.WaveformDataset or seisbench.data.MultiWaveformDataset
    :param metadata: The additional information as pandas dataframe.
                     Each row corresponds to one sample from the generator.
    :type metadata: pandas.DataFrame
    """

    def __init__(self, dataset, metadata):
        self.metadata = metadata
        super().__init__(dataset)

    def __str__(self):
        summary = f"SteeredGenerator with {len(self._augmentations)} augmentations:\n"
        for i, aug in enumerate(self._augmentations):
            summary += f" {i + 1}.\t{str(aug)}\n"
        return summary

    def __len__(self):
        return len(self.metadata)

    def __getitem__(self, idx):
        control = self.metadata.iloc[idx].to_dict()
        kwargs = {
            "trace_name": control["trace_name"],
            "chunk": control.get("trace_chunk", None),
            "dataset": control.get("trace_dataset", None),
        }
        data_idx = self.dataset.get_idx_from_trace_name(**kwargs)
        state_dict = {"X": self.dataset.get_sample(data_idx), "_control_": control}

        # Recursive application of augmentation processing methods
        for func in self._augmentations:
            func(state_dict)

        # Remove control information
        del state_dict["_control_"]
        # Remove all metadata from the output
        state_dict = {k: v[0] for k, v in state_dict.items()}

        return state_dict



from .base import WaveformModel

import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np
import scipy.signal
from scipy.interpolate import interp1d


class DeepDenoiser(WaveformModel):
    def __init__(self, sampling_rate=100, **kwargs):
        citation = (
            "Zhu, W., Mousavi, S. M., & Beroza, G. C. (2019). "
            "Seismic signal denoising and decomposition using deep neural networks. "
            "IEEE Transactions on Geoscience and Remote Sensing, 57.11(2019), 9476 - 9488. "
            "https://doi.org/10.1109/TGRS.2019.2926772"
        )

        super().__init__(
            citation=citation,
            in_samples=3000,
            output_type="array",
            default_args={"overlap": 1000},
            pred_sample=(0, 3000),
            labels=self.generate_label,
            sampling_rate=sampling_rate,
            grouping="channel",
            **kwargs,
        )

        self.inc = nn.Conv2d(2, 8, (3, 3), padding=(1, 1), bias=False)
        self.in_bn = nn.BatchNorm2d(8, eps=1e-3)

        self.down_conv_blocks = nn.ModuleList(
            [DownConvBlock(8 * 2 ** max(0, i - 1), 8 * 2**i) for i in range(5)]
        )

        self.conv5 = nn.Conv2d(128, 256, (3, 3), padding=(1, 1), bias=False)
        self.bn5 = nn.BatchNorm2d(256, eps=1e-3)

        self.up_conv_blocks = nn.ModuleList(
            [UpConvBlock(8 * 2 ** (5 - i), 8 * 2 ** (4 - i)) for i in range(5)]
        )

        self.outc = nn.Conv2d(8, 2, (1, 1), bias=True)

    def forward(self, x):
        x = torch.relu(self.in_bn(self.inc(x)))

        mids = []
        for layer in self.down_conv_blocks:
            x, mid = layer(x)
            mids.append(mid)

        x = torch.relu(self.bn5(self.conv5(x)))

        for layer, mid in zip(self.up_conv_blocks, mids[::-1]):
            x = layer(x, mid)

        logits = self.outc(x)
        preds = torch.softmax(logits, dim=1)

        return preds

    @staticmethod
    def generate_label(trace_stats):
        # Simply use channel as label
        return trace_stats.channel

    def annotate_window_pre(self, window, argdict):
        f, t, tmp_signal = scipy.signal.stft(
            window, fs=self.sampling_rate, nperseg=30, nfft=60, boundary="zeros"
        )
        noisy_signal = np.stack([tmp_signal.real, tmp_signal.imag], axis=0)

        noisy_signal[np.isnan(noisy_signal)] = 0
        noisy_signal[np.isinf(noisy_signal)] = 0

        return self._normalize_batch(noisy_signal), noisy_signal

    @staticmethod
    def _normalize_batch(data, window=200):
        """
        Adapted from original DeepDenoiser implementation available at
        https://github.com/wayneweiqiang/DeepDenoiser/blob/7bd9284ece73e25c99db2ad101aacda2a215a41a/deepdenoiser/app.py#L72

        data shape: 2, nf, nt
        data: nbn, nf, nt, 2
        """
        data = np.expand_dims(data, 0)  # 1 (nbt), 2, nf, nt
        data = data.transpose(0, 2, 3, 1)  # nbn, nf, nt, 2
        assert len(data.shape) == 4
        shift = window // 2
        nbt, nf, nt, nimg = data.shape

        # std in slide windows
        data_pad = np.pad(
            data, ((0, 0), (0, 0), (window // 2, window // 2), (0, 0)), mode="reflect"
        )
        t = np.arange(0, nt + shift - 1, shift, dtype="int")  # 201 => 0, 100, 200

        std = np.zeros([nbt, len(t)])
        mean = np.zeros([nbt, len(t)])
        for i in range(std.shape[1]):
            std[:, i] = np.std(
                data_pad[:, :, i * shift : i * shift + window, :], axis=(1, 2, 3)
            )
            mean[:, i] = np.mean(
                data_pad[:, :, i * shift : i * shift + window, :], axis=(1, 2, 3)
            )

        std[:, -1], mean[:, -1] = std[:, -2], mean[:, -2]
        std[:, 0], mean[:, 0] = std[:, 1], mean[:, 1]

        # normalize data with interplated std
        t_interp = np.arange(nt, dtype="int")
        std_interp = interp1d(t, std, kind="slinear")(t_interp)
        std_interp[std_interp == 0] = 1.0
        mean_interp = interp1d(t, mean, kind="slinear")(t_interp)

        data = (data - mean_interp[:, np.newaxis, :, np.newaxis]) / std_interp[
            :, np.newaxis, :, np.newaxis
        ]

        if len(t) > 3:  # need to address this normalization issue in training
            data /= 2.0

        data = data.transpose(0, 3, 1, 2)  # 1 (nbt), 2, nf, nt
        data = data[0]  # Remove batch dim

        return data

    def annotate_window_post(self, pred, piggyback=None, argdict=None):
        noisy_signal = piggyback
        _, denoised_signal = scipy.signal.istft(
            (noisy_signal[0] + noisy_signal[1] * 1j) * pred[0],
            fs=self.sampling_rate,
            nperseg=30,
            nfft=60,
            boundary="zeros",
        )
        return denoised_signal.reshape(-1, 1)

    def get_model_args(self):
        model_args = super().get_model_args()
        for key in [
            "citation",
            "in_samples",
            "output_type",
            "default_args",
            "pred_sample",
            "labels",
            "sampling_rate",
            "grouping",
        ]:
            del model_args[key]

        model_args["sampling_rate"] = self.sampling_rate

        return model_args


class DownConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()

        self.conv1 = nn.Conv2d(
            in_channels, out_channels, (3, 3), padding=(1, 1), bias=False
        )
        self.bn1 = nn.BatchNorm2d(out_channels, eps=1e-3)
        self.conv2 = nn.Conv2d(
            out_channels,
            out_channels,
            (3, 3),
            stride=(2, 2),
            padding=(1, 1),
            bias=False,
        )
        self.bn2 = nn.BatchNorm2d(out_channels, eps=1e-3)

    def forward(self, x):
        x = torch.relu(self.bn1(self.conv1(x)))
        mid = x

        # Required for compatibility with tensorflow version - stride is treated differently otherwise
        remove2 = False
        remove3 = False
        if x.shape[2] % 2 == 0:
            x = F.pad(x, (0, 0, 1, 0), "constant", 0)
            remove2 = True
        if x.shape[3] % 2 == 0:
            x = F.pad(x, (1, 0), "constant", 0)
            remove3 = True

        x = self.conv2(x)

        if remove2:
            x = x[:, :, 1:]  # Remove padding
        if remove3:
            x = x[:, :, :, 1:]  # Remove padding

        x = torch.relu(self.bn2(x))

        return x, mid


class UpConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()

        self.conv1 = nn.ConvTranspose2d(
            in_channels, out_channels, (3, 3), stride=(2, 2), padding=(0, 0), bias=False
        )
        self.bn1 = nn.BatchNorm2d(out_channels, eps=1e-3)

        # Again in_channels to account for the added residual connections
        self.conv2 = nn.Conv2d(
            in_channels, out_channels, (3, 3), padding=(1, 1), bias=False
        )
        self.bn2 = nn.BatchNorm2d(out_channels, eps=1e-3)

    def forward(self, x, mid):
        x = self.conv1(x)

        x = torch.relu(self.bn1(x))

        # Truncation is necessary to get correct shapes and be compatible with tensorflow implementation
        if mid.shape[2] % 2 == 0:
            x = x[:, :, :-1]
        else:
            x = x[:, :, :-2]

        if mid.shape[3] % 2 == 0:
            x = x[:, :, :, :-1]
        else:
            x = x[:, :, :, :-2]

        x = torch.cat([mid, x], dim=1)

        x = torch.relu(self.bn2(self.conv2(x)))

        return x



from .base import WaveformModel, ActivationLSTMCell, CustomLSTM

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np


# For implementation, potentially follow: https://medium.com/huggingface/from-tensorflow-to-pytorch-265f40ef2a28
class EQTransformer(WaveformModel):
    """
    The EQTranformer from Mousavi et al. (2020)

    Implementation adapted from the Github repository https://github.com/smousavi05/EQTransformer
    Assumes padding="same" and activation="relu" as in the pretrained EQTransformer models

    By instantiating the model with `from_pretrained("original")` a binary compatible version of the original
    EQTransformer with the original weights from Mousavi et al. (2020) can be loaded.

    :param in_channels: Number of input channels, by default 3.
    :param in_samples: Number of input samples per channel, by default 6000.
                       The model expects input shape (in_channels, in_samples)
    :param classes: Number of output classes, by default 2. The detection channel is not counted.
    :param phases: Phase hints for the classes, by default "PS". Can be None.
    :param res_cnn_blocks: Number of residual convolutional blocks
    :param lstm_blocks: Number of LSTM blocks
    :param drop_rate: Dropout rate
    :param original_compatible: If True, uses a few custom layers for binary compatibility with original model
                                from Mousavi et al. (2020).
                                This option defaults to False.
                                It is usually recommended to stick to the default value, as the custom layers show
                                slightly worse performance than the PyTorch builtins.
                                The exception is when loading the original weights using :py:func:`from_pretrained`.
    :param kwargs: Keyword arguments passed to the constructor of :py:class:`WaveformModel`.
    """

    def __init__(
        self,
        in_channels=3,
        in_samples=6000,
        classes=2,
        phases="PS",
        lstm_blocks=3,
        drop_rate=0.1,
        original_compatible=False,
        sampling_rate=100,
        **kwargs,
    ):
        citation = (
            "Mousavi, S.M., Ellsworth, W.L., Zhu, W., Chuang, L, Y., and Beroza, G, C. "
            "Earthquake transformer—an attentive deep-learning model for simultaneous earthquake "
            "detection and phase picking. Nat Commun 11, 3952 (2020). "
            "https://doi.org/10.1038/s41467-020-17591-w"
        )
        # Blinding defines how many samples at beginning and end of the prediction should be ignored
        # This is usually required to mitigate prediction problems from training properties, e.g.,
        # if all picks in the training fall between seconds 5 and 55.
        super().__init__(
            citation=citation,
            output_type="array",
            default_args={"overlap": 1800, "blinding": (500, 500)},
            in_samples=in_samples,
            pred_sample=(0, in_samples),
            labels=["Detection"] + list(phases),
            sampling_rate=sampling_rate,
            **kwargs,
        )

        self.in_channels = in_channels
        self.classes = classes
        self.lstm_blocks = lstm_blocks
        self.drop_rate = drop_rate
        self.original_compatible = original_compatible

        if original_compatible and in_samples != 6000:
            raise ValueError("original_compatible=True requires in_samples=6000.")

        self._phases = phases
        if phases is not None and len(phases) != classes:
            raise ValueError(
                f"Number of classes ({classes}) does not match number of phases ({len(phases)})."
            )

        # Parameters from EQTransformer repository
        self.filters = [
            8,
            16,
            16,
            32,
            32,
            64,
            64,
        ]  # Number of filters for the convolutions
        self.kernel_sizes = [11, 9, 7, 7, 5, 5, 3]  # Kernel sizes for the convolutions
        self.res_cnn_kernels = [3, 3, 3, 3, 2, 3, 2]

        # TODO: Add regularizers when training model
        # kernel_regularizer=keras.regularizers.l2(1e-6),
        # bias_regularizer=keras.regularizers.l1(1e-4),

        # Encoder stack
        self.encoder = Encoder(
            input_channels=self.in_channels,
            filters=self.filters,
            kernel_sizes=self.kernel_sizes,
            in_samples=self.in_samples,
        )

        # Res CNN Stack
        self.res_cnn_stack = ResCNNStack(
            kernel_sizes=self.res_cnn_kernels,
            filters=self.filters[-1],
            drop_rate=self.drop_rate,
        )

        # BiLSTM stack
        self.bi_lstm_stack = BiLSTMStack(
            blocks=self.lstm_blocks,
            input_size=self.filters[-1],
            drop_rate=self.drop_rate,
            original_compatible=original_compatible,
        )

        # Global attention - two transformers
        self.transformer_d0 = Transformer(input_size=16, drop_rate=self.drop_rate)
        self.transformer_d = Transformer(input_size=16, drop_rate=self.drop_rate)

        # Detection decoder and final Conv
        self.decoder_d = Decoder(
            input_channels=16,
            filters=self.filters[::-1],
            kernel_sizes=self.kernel_sizes[::-1],
            out_samples=in_samples,
            original_compatible=original_compatible,
        )
        self.conv_d = nn.Conv1d(
            in_channels=self.filters[0], out_channels=1, kernel_size=11, padding=5
        )

        # Picking branches
        self.pick_lstms = []
        self.pick_attentions = []
        self.pick_decoders = []
        self.pick_convs = []
        self.dropout = nn.Dropout(drop_rate)

        for _ in range(self.classes):
            if original_compatible:
                lstm = CustomLSTM(ActivationLSTMCell, 16, 16, bidirectional=False)
            else:
                lstm = nn.LSTM(16, 16, bidirectional=False)
            self.pick_lstms.append(lstm)

            attention = SeqSelfAttention(input_size=16, attention_width=3)
            self.pick_attentions.append(attention)

            decoder = Decoder(
                input_channels=16,
                filters=self.filters[::-1],
                kernel_sizes=self.kernel_sizes[::-1],
                out_samples=in_samples,
                original_compatible=original_compatible,
            )
            self.pick_decoders.append(decoder)

            conv = nn.Conv1d(
                in_channels=self.filters[0], out_channels=1, kernel_size=11, padding=5
            )
            self.pick_convs.append(conv)

        self.pick_lstms = nn.ModuleList(self.pick_lstms)
        self.pick_attentions = nn.ModuleList(self.pick_attentions)
        self.pick_decoders = nn.ModuleList(self.pick_decoders)
        self.pick_convs = nn.ModuleList(self.pick_convs)

    def forward(self, x):
        assert x.ndim == 3
        assert x.shape[1:] == (self.in_channels, self.in_samples)

        # Shared encoder part
        x = self.encoder(x)
        x = self.res_cnn_stack(x)
        x = self.bi_lstm_stack(x)
        x, _ = self.transformer_d0(x)
        x, _ = self.transformer_d(x)

        # Detection part
        detection = self.decoder_d(x)
        detection = torch.sigmoid(self.conv_d(detection))
        detection = torch.squeeze(detection, dim=1)  # Remove channel dimension

        outputs = [detection]

        # Pick parts
        for lstm, attention, decoder, conv in zip(
            self.pick_lstms, self.pick_attentions, self.pick_decoders, self.pick_convs
        ):
            px = x.permute(
                2, 0, 1
            )  # From batch, channels, sequence to sequence, batch, channels
            px = lstm(px)[0]
            px = self.dropout(px)
            px = px.permute(
                1, 2, 0
            )  # From sequence, batch, channels to batch, channels, sequence
            px, _ = attention(px)
            px = decoder(px)
            pred = torch.sigmoid(conv(px))
            pred = torch.squeeze(pred, dim=1)  # Remove channel dimension

            outputs.append(pred)

        return tuple(outputs)

    def annotate_window_post(self, pred, piggyback=None, argdict=None):
        # Combine predictions in one array
        prenan, postnan = argdict.get("blinding", (0, 0))
        pred = np.stack(pred, axis=-1)
        if prenan > 0:
            pred[:prenan] = np.nan
        if postnan > 0:
            pred[-postnan:] = np.nan
        return pred

    def annotate_window_pre(self, window, argdict):
        # Add a demean and an amplitude normalization step to the preprocessing
        window = window - np.mean(window, axis=-1, keepdims=True)
        window = window / (np.std(window) + 1e-10)

        # Cosine taper (very short, i.e., only six samples on each side)
        tap = 0.5 * (1 + np.cos(np.linspace(np.pi, 2 * np.pi, 6)))
        window[:, :6] *= tap
        window[:, -6:] *= tap[::-1]

        return window

    @property
    def phases(self):
        if self._phases is not None:
            return self._phases
        else:
            return list(range(self.classes))

    def classify_aggregate(self, annotations, argdict):
        """
        Converts the annotations to discrete picks using
        :py:func:`~seisbench.models.base.WaveformModel.picks_from_annotations`
        and to discrete detections using :py:func:`~seisbench.models.base.WaveformModel.detections_from_annotations`.
        Trigger onset thresholds for picks are derived from the argdict at keys "[phase]_threshold".
        Trigger onset thresholds for detections are derived from the argdict at key "detection_threshold".

        :param annotations: See description in superclass
        :param argdict: See description in superclass
        :return: List of picks, list of detections
        """
        picks = []
        for phase in self.phases:
            picks += self.picks_from_annotations(
                annotations.select(channel=f"EQTransformer_{phase}"),
                argdict.get(f"{phase}_threshold", 0.1),
                phase,
            )

        detections = self.detections_from_annotations(
            annotations.select(channel="EQTransformer_Detection"),
            argdict.get("detection_threshold", 0.3),
        )

        return sorted(picks), sorted(detections)

    def get_model_args(self):
        model_args = super().get_model_args()
        for key in [
            "citation",
            "in_samples",
            "output_type",
            "default_args",
            "pred_sample",
            "labels",
            "sampling_rate",
        ]:
            del model_args[key]

        model_args["in_channels"] = self.in_channels
        model_args["in_samples"] = self.in_samples
        model_args["classes"] = self.classes
        model_args["phases"] = self.phases
        model_args["lstm_blocks"] = self.lstm_blocks
        model_args["drop_rate"] = self.drop_rate
        model_args["original_compatible"] = self.original_compatible
        model_args["sampling_rate"] = self.sampling_rate

        return model_args


class Encoder(nn.Module):
    """
    Encoder stack
    """

    def __init__(self, input_channels, filters, kernel_sizes, in_samples):
        super().__init__()

        convs = []
        pools = []
        self.paddings = []
        for in_channels, out_channels, kernel_size in zip(
            [input_channels] + filters[:-1], filters, kernel_sizes
        ):
            convs.append(
                nn.Conv1d(
                    in_channels, out_channels, kernel_size, padding=kernel_size // 2
                )
            )

            # To be consistent with the behaviour in tensorflow,
            # padding needs to be added for odd numbers of input_samples
            padding = in_samples % 2

            # Padding for MaxPool1d needs to be handled manually to conform with tf padding
            self.paddings.append(padding)
            pools.append(nn.MaxPool1d(2, padding=0))
            in_samples = (in_samples + padding) // 2

        self.convs = nn.ModuleList(convs)
        self.pools = nn.ModuleList(pools)

    def forward(self, x):
        for conv, pool, padding in zip(self.convs, self.pools, self.paddings):
            x = torch.relu(conv(x))
            if padding != 0:
                # Only pad right, use -1e10 as negative infinity
                x = F.pad(x, (0, padding), "constant", -1e10)
            x = pool(x)

        return x


class Decoder(nn.Module):
    def __init__(
        self,
        input_channels,
        filters,
        kernel_sizes,
        out_samples,
        original_compatible=False,
    ):
        super().__init__()

        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")
        self.original_compatible = original_compatible

        # We need to trim of the final sample sometimes to get to the right number of output samples
        self.crops = []
        current_samples = out_samples
        for i, _ in enumerate(filters):
            padding = current_samples % 2
            current_samples = (current_samples + padding) // 2
            if padding == 1:
                self.crops.append(len(filters) - 1 - i)

        convs = []
        for in_channels, out_channels, kernel_size in zip(
            [input_channels] + filters[:-1], filters, kernel_sizes
        ):
            convs.append(
                nn.Conv1d(
                    in_channels, out_channels, kernel_size, padding=kernel_size // 2
                )
            )

        self.convs = nn.ModuleList(convs)

    def forward(self, x):
        for i, conv in enumerate(self.convs):
            x = self.upsample(x)

            if self.original_compatible:
                if i == 3:
                    x = x[:, :, 1:-1]
            else:
                if i in self.crops:
                    x = x[:, :, :-1]

            x = F.relu(conv(x))

        return x


class ResCNNStack(nn.Module):
    def __init__(self, kernel_sizes, filters, drop_rate):
        super().__init__()

        members = []
        for ker in kernel_sizes:
            members.append(ResCNNBlock(filters, ker, drop_rate))

        self.members = nn.ModuleList(members)

    def forward(self, x):
        for member in self.members:
            x = member(x)

        return x


class ResCNNBlock(nn.Module):
    def __init__(self, filters, ker, drop_rate):
        super().__init__()

        self.manual_padding = False
        if ker == 3:
            padding = 1
        else:
            # ker == 2
            # Manual padding emulate the padding in tensorflow
            self.manual_padding = True
            padding = 0

        self.dropout = SpatialDropout1d(drop_rate)

        self.norm1 = nn.BatchNorm1d(filters, eps=1e-3)
        self.conv1 = nn.Conv1d(filters, filters, ker, padding=padding)

        self.norm2 = nn.BatchNorm1d(filters, eps=1e-3)
        self.conv2 = nn.Conv1d(filters, filters, ker, padding=padding)

    def forward(self, x):
        y = self.norm1(x)
        y = F.relu(y)
        y = self.dropout(y)
        if self.manual_padding:
            y = F.pad(y, (0, 1), "constant", 0)
        y = self.conv1(y)

        y = self.norm2(y)
        y = F.relu(y)
        y = self.dropout(y)
        if self.manual_padding:
            y = F.pad(y, (0, 1), "constant", 0)
        y = self.conv2(y)

        return x + y


class BiLSTMStack(nn.Module):
    def __init__(
        self, blocks, input_size, drop_rate, hidden_size=16, original_compatible=False
    ):
        super().__init__()

        # First LSTM has a different input size as the subsequent ones
        self.members = nn.ModuleList(
            [
                BiLSTMBlock(
                    input_size,
                    hidden_size,
                    drop_rate,
                    original_compatible=original_compatible,
                )
            ]
            + [
                BiLSTMBlock(
                    hidden_size,
                    hidden_size,
                    drop_rate,
                    original_compatible=original_compatible,
                )
                for _ in range(blocks - 1)
            ]
        )

    def forward(self, x):
        for member in self.members:
            x = member(x)
        return x


class BiLSTMBlock(nn.Module):
    def __init__(self, input_size, hidden_size, drop_rate, original_compatible=False):
        super().__init__()

        if original_compatible:
            self.lstm = CustomLSTM(ActivationLSTMCell, input_size, hidden_size)
        else:
            self.lstm = nn.LSTM(input_size, hidden_size, bidirectional=True)
        self.dropout = nn.Dropout(drop_rate)
        self.conv = nn.Conv1d(2 * hidden_size, hidden_size, 1)
        self.norm = nn.BatchNorm1d(hidden_size, eps=1e-3)

    def forward(self, x):
        x = x.permute(
            2, 0, 1
        )  # From batch, channels, sequence to sequence, batch, channels
        x = self.lstm(x)[0]
        x = self.dropout(x)
        x = x.permute(
            1, 2, 0
        )  # From sequence, batch, channels to batch, channels, sequence
        x = self.conv(x)
        x = self.norm(x)
        return x


class Transformer(nn.Module):
    def __init__(self, input_size, drop_rate, attention_width=None):
        super().__init__()

        self.attention = SeqSelfAttention(input_size, attention_width=attention_width)
        self.norm1 = LayerNormalization(input_size)
        self.ff = FeedForward(input_size, drop_rate)
        self.norm2 = LayerNormalization(input_size)

    def forward(self, x):
        y, weight = self.attention(x)
        y = x + y
        y = self.norm1(y)
        y2 = self.ff(y)
        y2 = y + y2
        y2 = self.norm2(y2)

        return y2, weight


class SeqSelfAttention(nn.Module):
    """
    Additive self attention
    """

    def __init__(self, input_size, units=32, attention_width=None, eps=1e-5):
        super().__init__()
        self.attention_width = attention_width

        self.Wx = nn.Parameter(uniform(-0.02, 0.02, input_size, units))
        self.Wt = nn.Parameter(uniform(-0.02, 0.02, input_size, units))
        self.bh = nn.Parameter(torch.zeros(units))

        self.Wa = nn.Parameter(uniform(-0.02, 0.02, units, 1))
        self.ba = nn.Parameter(torch.zeros(1))

        self.eps = eps

    def forward(self, x):
        # x.shape == (batch, channels, time)

        x = x.permute(0, 2, 1)  # to (batch, time, channels)

        q = torch.unsqueeze(
            torch.matmul(x, self.Wt), 2
        )  # Shape (batch, time, 1, channels)
        k = torch.unsqueeze(
            torch.matmul(x, self.Wx), 1
        )  # Shape (batch, 1, time, channels)

        h = torch.tanh(q + k + self.bh)

        # Emissions
        e = torch.squeeze(
            torch.matmul(h, self.Wa) + self.ba, -1
        )  # Shape (batch, time, time)

        if self.attention_width is None:
            a = F.softmax(e, dim=-1)
        else:
            e = e - torch.max(x, dim=-1, keepdim=True)[0]
            e = torch.exp(e)
            lower = (
                torch.arange(0, e.shape[1], device=e.device) - self.attention_width // 2
            )
            upper = lower + self.attention_width
            indices = torch.unsqueeze(torch.arange(0, e.shape[1], device=e.device), 1)
            mask = torch.logical_and(lower <= indices, indices < upper)
            e = torch.where(mask, e, torch.zeros_like(e))
            a = e / (torch.sum(e, dim=-1, keepdim=True) + self.eps)

        v = torch.matmul(a, x)

        v = v.permute(0, 2, 1)  # to (batch, channels, time)

        return v, a


def uniform(a, b, *args):
    return a + (b - a) * torch.rand(*args)


class LayerNormalization(nn.Module):
    def __init__(self, filters, eps=1e-14):
        super().__init__()

        gamma = torch.ones(filters, 1)
        self.gamma = nn.Parameter(gamma)
        beta = torch.zeros(filters, 1)
        self.beta = nn.Parameter(beta)
        self.eps = eps

    def forward(self, x):
        mean = torch.mean(x, 1, keepdim=True)
        var = torch.mean((x - mean) ** 2, 1, keepdim=True) + self.eps
        std = torch.sqrt(var)
        outputs = (x - mean) / std

        outputs = outputs * self.gamma
        outputs = outputs + self.beta

        return outputs


class FeedForward(nn.Module):
    def __init__(self, io_size, drop_rate, hidden_size=128):
        super().__init__()

        self.lin1 = nn.Linear(io_size, hidden_size)
        self.lin2 = nn.Linear(hidden_size, io_size)
        self.dropout = nn.Dropout(drop_rate)

    def forward(self, x):
        x = x.permute(0, 2, 1)  # To (batch, time, channel)
        x = F.relu(self.lin1(x))
        x = self.dropout(x)
        x = self.lin2(x)
        x = x.permute(0, 2, 1)  # To (batch, channel, time)

        return x


class SpatialDropout1d(nn.Module):
    def __init__(self, drop_rate):
        super().__init__()

        self.drop_rate = drop_rate
        self.dropout = nn.Dropout2d(drop_rate)

    def forward(self, x):
        x = x.unsqueeze(dim=-1)  # Add fake dimension
        x = self.dropout(x)
        x = x.squeeze(dim=-1)  # Remove fake dimension
        return x



import seisbench
import functools


def log_lifecycle(level):
    """
    Logs the invocation and termination of a function to seisbench.logger. Should be used as a decorator.
    """

    def decorator(func):
        @functools.wraps(func)
        def f(*args, **kwargs):
            seisbench.logger.log(msg=f"Starting {func.__name__}", level=level)
            res = func(*args, **kwargs)
            seisbench.logger.log(msg=f"Stopping {func.__name__}", level=level)
            return res

        return f

    return decorator



import seisbench
import seisbench.util
from .base import BenchmarkDataset

import h5py
from collections import defaultdict
import numpy as np
import shutil


class SCEDC(BenchmarkDataset):
    """
    SCEDC waveform archive (2000-2020).

    Splits are set using standard random sampling of :py:class: BenchmarkDataset.
    """

    def __init__(self, **kwargs):
        citation = (
            "SCEDC (2013): Southern California Earthquake Center."
            "https://doi.org/10.7909/C3WD3xH1"
        )

        seisbench.logger.warning(
            "Check available storage and memory before downloading and general use "
            "of SCEDC dataset. "
            "Dataset size: waveforms.hdf5 ~660Gb, metadata.csv ~2.2Gb"
        )

        super().__init__(citation=citation, repository_lookup=True, **kwargs)

    def _download_dataset(self, **kwargs):
        # NOTE: SCEDC dataset is pre-compiled and stored in remote repository root for access
        pass


# TODO: Check with Zach Ross if this dataset really only differs from Ross2018JGRPick through the class rebalancing.
#       If so, it this should be stated in the SeisBench documentation and probably also be reflected in the naming.
class Ross2018JGRFM(BenchmarkDataset):
    """
    First motion polarity dataset belonging to the publication:
    Ross, Z. E., Meier, M.‐A., & Hauksson, E. (2018). P wave arrival picking and first‐motion polarity determination
    with deep learning. Journal of Geophysical Research: Solid Earth, 123, 5120– 5129.
    https://doi.org/10.1029/2017JB015251

    Note that this dataset contains picks as well.

    .. warning::

        This dataset only contains traces for the Z component.
        It therefore ignores the default SeisBench the component_order.

    """

    def __init__(self, **kwargs):
        citation = (
            "Ross, Z. E., Meier, M.‐A., & Hauksson, E. (2018). "
            "P wave arrival picking and first‐motion polarity determination with deep learning. "
            "Journal of Geophysical Research: Solid Earth, 123, 5120– 5129. https://doi.org/10.1029/2017JB015251"
        )
        super().__init__(
            citation=citation, repository_lookup=False, component_order="Z", **kwargs
        )

    def _download_dataset(self, writer, cleanup=False, blocksize=2**14):
        """
        Downloads and converts the dataset from the original publication

        :param writer: WaveformWriter
        :param cleanup: If true, delete the original files after conversion. Defaults to false.
        :param blocksize: Number of waveform samples to read from disk at once.
        :return:
        """

        path = self.path
        path_original = path / "original"
        path_original.mkdir(parents=True, exist_ok=True)

        # Maps ids to strings for the polarity
        polarity_list = ["up", "down", "unknown"]

        # Download data files
        data_urls = [
            "https://service.scedc.caltech.edu/ftp/Ross_FinalTrainedModels/scsn_p_2000_2017_6sec_0.5r_fm_test.hdf5",
            "https://service.scedc.caltech.edu/ftp/Ross_FinalTrainedModels/scsn_p_2000_2017_6sec_0.5r_fm_train.hdf5",
        ]

        for f in data_urls:
            # Uses callback_if_uncached only to be able to utilize the cache mechanism.
            # Concurrent accesses are anyhow already controlled
            # by the callback_if_uncached call wrapping _download_dataset.
            # It's therefore considered save to set force=True.
            filename = f[f.rfind("/") + 1 :]

            def callback_download_original(path):
                seisbench.util.download_http(
                    f,
                    path,
                    desc=f"Downloading file {filename}",
                )

            seisbench.util.callback_if_uncached(
                path_original / filename, callback_download_original, force=True
            )

        with h5py.File(
            path_original / "scsn_p_2000_2017_6sec_0.5r_fm_train.hdf5", "r"
        ) as f_train:
            train_samples = f_train["X"].shape[0]
        with h5py.File(
            path_original / "scsn_p_2000_2017_6sec_0.5r_fm_test.hdf5", "r"
        ) as f_test:
            test_samples = f_test["X"].shape[0]

        writer.set_total(train_samples + test_samples)
        writer.data_format = {
            "dimension_order": "CW",
            "component_order": "Z",
            "measurement": "velocity",
            "sampling_rate": 100,
            "unit": "none/normalized",
            "instrument_response": "not restituted",
        }

        eq_counts = defaultdict(lambda: 0)

        for split in ["train", "test"]:
            with h5py.File(
                path_original / f"scsn_p_2000_2017_6sec_0.5r_fm_{split}.hdf5", "r"
            ) as f:
                # Preload all small arrays to avoid disk seeks
                y = f["Y"][:]
                dist = f["dist"][:]
                evids = f["evids"][:]
                mag = f["mag"][:]
                sncls = f["sncls"][:]
                snr = f["snr"][:]

                # Use 10 percent of the training events as development set
                if split == "train":
                    dev_ids = set(np.unique(evids)[::10])
                else:
                    dev_ids = set()

                wf_block = None
                for i in range(f["X"].shape[0]):
                    # Preload block of waveforms
                    if i % blocksize == 0:
                        wf_block = f["X"][i : i + blocksize]
                    wf = wf_block[i % blocksize].reshape(
                        1, -1
                    )  # Load waveforms and add (virtual) channel axis

                    eid = f"{evids[i]}_{sncls[i].decode()}"
                    trace_station_id = eq_counts[eid]
                    eq_counts[eid] += 1
                    trace_name = f"{eid}_{trace_station_id}"

                    if evids[i] in dev_ids:
                        trace_split = "dev"
                    else:
                        trace_split = split

                    net, sta, cha = sncls[i].decode().split(".")
                    polarity = polarity_list[y[i]]

                    metadata = {
                        "trace_name": trace_name,
                        "trace_category": "earthquake",
                        "trace_p_arrival_sample": 300,
                        "trace_p_status": "manual",
                        "trace_snr_db": snr[i],
                        "trace_channel": cha,
                        "trace_polarity": polarity,
                        "station_network_code": net,
                        "station_code": sta,
                        "source_magnitude": mag[i],
                        "source_id": evids[i],
                        "path_ep_distance_km": dist[i],
                        "split": trace_split,
                    }

                    writer.add_trace(metadata, wf)

            # Write out all data from the current split
            writer.flush_hdf5()

        if cleanup:
            shutil.rmtree(path_original)


class Ross2018JGRPick(BenchmarkDataset):
    """
    Pick dataset belonging to the publication:
    Ross, Z. E., Meier, M.‐A., & Hauksson, E. (2018). P wave arrival picking and first‐motion polarity determination
    with deep learning. Journal of Geophysical Research: Solid Earth, 123, 5120– 5129.
    https://doi.org/10.1029/2017JB015251

    Note that this dataset contains polarities as well.

    .. warning::

        This dataset only contains traces for the Z component.
        It therefore ignores the default SeisBench the component_order.

    """

    def __init__(self, **kwargs):
        citation = (
            "Ross, Z. E., Meier, M.‐A., & Hauksson, E. (2018). "
            "P wave arrival picking and first‐motion polarity determination with deep learning. "
            "Journal of Geophysical Research: Solid Earth, 123, 5120– 5129. https://doi.org/10.1029/2017JB015251"
        )
        super().__init__(
            citation=citation, repository_lookup=False, component_order="Z", **kwargs
        )

    def _download_dataset(self, writer, cleanup=False, blocksize=2**14):
        """
        Downloads and converts the dataset from the original publication

        :param writer: WaveformWriter
        :param cleanup: If true, delete the original files after conversion. Defaults to false.
        :param blocksize: Number of waveform samples to read from disk at once
        :return:
        """

        path = self.path
        path_original = path / "original"
        path_original.mkdir(parents=True, exist_ok=True)

        # Maps ids to strings for the polarity
        polarity_list = ["up", "down", "unknown"]

        # Download data files
        data_urls = [
            "https://service.scedc.caltech.edu/ftp/Ross_FinalTrainedModels/scsn_p_2000_2017_6sec_0.5r_pick_test.hdf5",
            "https://service.scedc.caltech.edu/ftp/Ross_FinalTrainedModels/scsn_p_2000_2017_6sec_0.5r_pick_train.hdf5",
        ]

        for f in data_urls:
            # Uses callback_if_uncached only to be able to utilize the cache mechanism.
            # Concurrent accesses are anyhow already controlled
            # by the callback_if_uncached call wrapping _download_dataset.
            # It's therefore considered save to set force=True.
            filename = f[f.rfind("/") + 1 :]

            def callback_download_original(path):
                seisbench.util.download_http(
                    f,
                    path,
                    desc=f"Downloading file {filename}",
                )

            seisbench.util.callback_if_uncached(
                path_original / filename, callback_download_original, force=True
            )

        with h5py.File(
            path_original / "scsn_p_2000_2017_6sec_0.5r_pick_train.hdf5", "r"
        ) as f_train:
            train_samples = f_train["X"].shape[0]
        with h5py.File(
            path_original / "scsn_p_2000_2017_6sec_0.5r_pick_test.hdf5", "r"
        ) as f_test:
            test_samples = f_test["X"].shape[0]

        writer.set_total(train_samples + test_samples)
        writer.data_format = {
            "dimension_order": "CW",
            "component_order": "Z",
            "measurement": "velocity",
            "sampling_rate": 100,
            "unit": "none/normalized",
            "instrument_response": "not restituted",
        }

        eq_counts = defaultdict(lambda: 0)

        for split in ["train", "test"]:
            with h5py.File(
                path_original / f"scsn_p_2000_2017_6sec_0.5r_pick_{split}.hdf5", "r"
            ) as f:
                # Preload all small arrays to avoid disk seeks
                fm = f["fm"][:]
                dist = f["dist"][:]
                evids = f["evids"][:]
                mag = f["mag"][:]
                sncls = f["sncls"][:]
                snr = f["snr"][:]

                # Use 10 percent of the training events as development set
                if split == "train":
                    dev_ids = set(np.unique(evids)[::10])
                else:
                    dev_ids = set()

                wf_block = None
                for i in range(f["X"].shape[0]):
                    # Preload block of waveforms
                    if i % blocksize == 0:
                        wf_block = f["X"][i : i + blocksize]
                    wf = wf_block[i % blocksize].reshape(
                        1, -1
                    )  # Load waveforms and add (virtual) channel axis

                    eid = f"{evids[i]}_{sncls[i].decode()}"
                    trace_station_id = eq_counts[eid]
                    eq_counts[eid] += 1
                    trace_name = f"{eid}_{trace_station_id}"

                    if evids[i] in dev_ids:
                        trace_split = "dev"
                    else:
                        trace_split = split

                    net, sta, cha = sncls[i].decode().split(".")
                    polarity = polarity_list[fm[i]]

                    metadata = {
                        "trace_name": trace_name,
                        "trace_category": "earthquake",
                        "trace_p_arrival_sample": 300,
                        "trace_p_status": "manual",
                        "trace_snr_db": snr[i],
                        "trace_channel": cha,
                        "trace_polarity": polarity,
                        "station_network_code": net,
                        "station_code": sta,
                        "source_magnitude": mag[i],
                        "source_id": evids[i],
                        "path_ep_distance_km": dist[i],
                        "split": trace_split,
                    }

                    writer.add_trace(metadata, wf)

            # Write out all data from the current split
            writer.flush_hdf5()

        if cleanup:
            shutil.rmtree(path_original)


class Ross2018GPD(BenchmarkDataset):
    """
    Pick dataset belonging to the publication:
    Zachary E. Ross, Men‐Andrin Meier, Egill Hauksson, Thomas H. Heaton;
    Generalized Seismic Phase Detection with Deep Learning.
    Bulletin of the Seismological Society of America 2018;; 108 (5A): 2894–2901.
    https://doi.org/10.1785/0120180080
    """

    def __init__(self, **kwargs):
        citation = (
            "Ross, Z. E., Meier, M.‐A., Hauksson, E., & Heaton, T.(2018). "
            "Generalized Seismic Phase Detection with Deep Learning. "
            "Bulletin of the Seismological Society of America 2018;; 108 (5A): 2894–2901. "
            "https://doi.org/10.1785/0120180080"
        )
        super().__init__(citation=citation, repository_lookup=False, **kwargs)

    def _download_dataset(self, writer, cleanup=False, blocksize=2**14):
        """
        Downloads and converts the dataset from the original publication

        :param writer: WaveformWriter
        :param cleanup: If true, delete the original files after conversion. Defaults to false.
        :param blocksize: Number of waveform samples to read from disk at once
        :return:
        """

        path = self.path
        path_original = path / "original"
        path_original.mkdir(parents=True, exist_ok=True)

        # Download data files
        # Uses callback_if_uncached only to be able to utilize the cache mechanism
        # Concurrent accesses are anyhow already controlled by the callback_if_uncached call wrapping _download_dataset
        # It's therefore considered save to set force=True

        data_url = "https://service.scedc.caltech.edu/ftp/ross_etal_2018_bssa/scsn_ps_2000_2017_shuf.hdf5"
        filename = data_url[data_url.rfind("/") + 1 :]

        def callback_download_original(path):
            seisbench.util.download_http(
                data_url,
                path,
                desc=f"Downloading file {filename}",
            )

        seisbench.util.callback_if_uncached(
            path_original / filename, callback_download_original, force=True
        )

        writer.bucket_size = (
            4096  # Single waveforms are small so the bucket size should be larger
        )
        writer.data_format = {
            "dimension_order": "CW",
            "component_order": "ZNE",
            "measurement": "velocity",
            "sampling_rate": 100,
            "unit": "none/normalized",
            "instrument_response": "not restituted",
        }

        with h5py.File(path_original / filename, "r") as fin:
            writer.set_total(fin["X"].shape[0])
            y = fin["Y"][()]

            wf_block = None
            for i in range(fin["X"].shape[0]):
                # Preload block of waveforms
                if i % blocksize == 0:
                    wf_block = fin["X"][i : i + blocksize]
                wf = wf_block[i % blocksize].T  # Load waveforms and transpose
                wf = wf[[2, 0, 1]]  # Resort components to ZNE

                if i % 10 < 6:
                    trace_split = "train"
                elif i % 10 < 7:
                    trace_split = "dev"
                else:
                    trace_split = "test"

                metadata = {
                    "split": trace_split,
                }

                if y[i] == 0:
                    # P pick
                    metadata["trace_category"] = "earthquake"
                    metadata["trace_p_arrival_sample"] = 300
                    metadata["trace_p_status"] = "manual"
                elif y[i] == 1:
                    # S pick
                    metadata["trace_category"] = "earthquake"
                    metadata["trace_s_arrival_sample"] = 300
                    metadata["trace_s_status"] = "manual"
                else:
                    metadata["trace_category"] = "noise"

                writer.add_trace(metadata, wf)

        if cleanup:
            shutil.rmtree(path_original)


# TODO: Write Men-Andrin Meier regarding zero metadata columns, time format, split format
class Meier2019JGR(BenchmarkDataset):
    """
    Southern californian part of the dataset from Meier et al. (2019)
    Note that due to the missing Japanese data,
    there is a massive overrepresentation of noise samples.

    Meier, M.-A., Ross, Z. E., Ramachandran, A., Balakrishna, A.,
    Nair, S., Kundzicz, P., et al. (2019). Reliable real‐time
    seismic signal/noise discrimination with machine learning.
    Journal of Geophysical Research: Solid Earth, 124.
    https://doi.org/10.1029/2018JB016661
    """

    def __init__(self, **kwargs):
        citation = (
            "Meier, M.-A., Ross, Z. E., Ramachandran, A., Balakrishna, A., "
            "Nair, S., Kundzicz, P., et al. (2019). Reliable real‐time "
            "seismic signal/noise discrimination with machine learning. "
            "Journal of Geophysical Research: Solid Earth, 124. "
            "https://doi.org/10.1029/2018JB016661"
        )
        super().__init__(citation=citation, repository_lookup=False, **kwargs)

    def _download_dataset(self, writer, cleanup=False, blocksize=2**14):
        """
        Downloads and converts the dataset from the original publication

        :param writer: WaveformWriter
        :param cleanup: If true, delete the original files after conversion. Defaults to false.
        :param blocksize: Number of waveform samples to read from disk at once
        :return:
        """

        path = self.path
        path_original = path / "original"
        path_original.mkdir(parents=True, exist_ok=True)

        # Download data files
        # Uses callback_if_uncached only to be able to utilize the cache mechanism
        # Concurrent accesses are anyhow already controlled by the callback_if_uncached call wrapping _download_dataset
        # It's therefore considered save to set force=True

        data_url = "https://service.scedc.caltech.edu/ftp/meier_etal_2019_jgr/onsetWforms_meier19jgr_pub1_0_woJP.h5"
        filename = data_url[data_url.rfind("/") + 1 :]

        def callback_download_original(path):
            seisbench.util.download_http(
                data_url,
                path,
                desc=f"Downloading file {filename}",
            )

        seisbench.util.callback_if_uncached(
            path_original / filename, callback_download_original, force=True
        )

        writer.bucket_size = (
            4096  # Single waveforms are small so the bucket size should be larger
        )
        writer.data_format = {
            "dimension_order": "CW",
            "component_order": "ZNE",
            "measurement": "velocity",
            "sampling_rate": 100,
            "unit": "mps",
            "instrument_response": "gain corrected",
        }

        category_map = {
            "noise": "noise",
            "quake": "earthquake (local)",
            "tele": "earthquake (teleseismic)",
        }

        with h5py.File(path_original / filename, "r") as fin:
            total = (
                fin["quake/wforms"].shape[1]
                + fin["noise/wforms"].shape[1]
                + fin["tele/wforms"].shape[1]
            )
            writer.set_total(total)

            for group in "quake", "noise", "tele":
                gin = fin[group]
                meta_features = gin["numMeta"][()]
                category = category_map[group]

                wf_block = None
                for i in range(meta_features.shape[1]):
                    # Preload block of waveforms
                    if i % blocksize == 0:
                        wf_block = gin["wforms"][:, i : i + blocksize]
                    wf = wf_block[:, i % blocksize]  # Load waveforms
                    wf = wf[[2, 0, 1]]  # Resort components to ZNE

                    # TODO: Read/define split
                    meta_row = meta_features[:, i]

                    if group == "noise":
                        metadata = {
                            "trace_category": category,
                            # "split": trace_split,
                            "trace_snr_db": meta_row[3],
                            "trace_record_id": meta_row[4],
                            # meta_row[5] - pickIndex - ignored
                            # Data is consistently zero
                            # "station_latitude_deg": meta_row[6],
                            # "station_longitude_deg": meta_row[7],
                            # "trace_pga_mps2": meta_row[8],
                            # "trace_pgv_mps": meta_row[9],
                            # "trace_pgd_m": meta_row[10],
                            "source_origin_time": meta_row[11],  # Format unclear
                            # Data is consistently zero
                            # "path_back_azimuth_deg": meta_row[12]
                        }

                    else:
                        metadata = {
                            "trace_category": category,
                            # "split": trace_split,
                            "source_magnitude": meta_row[0],
                            "path_hyp_distance_km": meta_row[1],
                            "source_depth_km": meta_row[2],
                            "trace_snr_db": meta_row[3],
                            "trace_record_id": meta_row[4],
                            # meta_row[5] - pickIndex - ignored
                            "station_latitude_deg": meta_row[6],
                            "station_longitude_deg": meta_row[7],
                            # Data is consistently zero
                            # "trace_pga_mps2": meta_row[8],
                            # "trace_pgv_mps": meta_row[9],
                            # "trace_pgd_m": meta_row[10],
                            "source_origin_time": meta_row[11],  # Format unclear
                            # Data is consistently zero
                            # "path_back_azimuth_deg": meta_row[12]
                            "trace_p_arrival_sample": 201,
                            "trace_p_status": "manual",
                        }

                    writer.add_trace(metadata, wf)

        if cleanup:
            shutil.rmtree(path_original)



from .generator import GenericGenerator, SteeredGenerator
from .augmentation import (
    Normalize,
    Filter,
    FilterKeys,
    ChangeDtype,
    OneOf,
    NullAugmentation,
    ChannelDropout,
    AddGap,
    RandomArrayRotation,
    GaussianNoise,
    Copy,
)
from .labeling import (
    SupervisedLabeller,
    PickLabeller,
    ProbabilisticLabeller,
    DetectionLabeller,
    StandardLabeller,
    ProbabilisticPointLabeller,
    StepLabeller,
)
from .windows import (
    FixedWindow,
    SlidingWindow,
    WindowAroundSample,
    RandomWindow,
    SteeredWindow,
)



import seisbench
from .base import BenchmarkDataset
from seisbench.util.trace_ops import (
    rotate_stream_to_zne,
    stream_to_array,
    trace_has_spikes,
)

from pathlib import Path
import obspy
from obspy.io.mseed import InternalMSEEDError
from obspy.geodetics import gps2dist_azimuth
import copy
from collections import defaultdict
import numpy as np
import pandas as pd


class GEOFON(BenchmarkDataset):
    """
    GEOFON dataset consisting of both regional and teleseismic picks. Mostly contains P arrivals,
    but a few S arrivals are annotated as well. Contains data from 2010-2013. The dataset will be
    downloaded from the SeisBench repository on first usage.
    """

    def __init__(self, **kwargs):
        # TODO: Add citation
        citation = "GEOFON dataset"
        super().__init__(citation=citation, repository_lookup=True, **kwargs)

    def _download_dataset(
        self, writer, basepath=None, time_before=60, time_after=60, **kwargs
    ):
        if basepath is None:
            raise ValueError(
                "'basepath' needs to be set in the download_kwargs to start dataset conversion for source. "
                "If you are seeing this error, the SeisBench remote root might be unavailable. "
                "If it is available and you are still seeing this error, "
                "please get in touch with the developers."
            )

        component_order = "ZNE"

        writer.data_format = {
            "dimension_order": "CW",
            "component_order": component_order,
            "measurement": "velocity",
            "unit": "counts",
            "instrument_response": "not restituted",
        }

        basepath = Path(basepath)

        location_helper = LocationHelper(basepath)
        inventory = obspy.read_inventory(
            str(basepath / "inventory_without_response.xml")
        )

        for event_path in sorted(basepath.iterdir()):
            if not event_path.is_dir():
                continue
            quakeml = [
                x for x in event_path.iterdir() if x.name.endswith("preferred-only.xml")
            ]
            if len(quakeml) != 1:
                continue

            quakeml = quakeml[0]

            catalog = obspy.read_events(str(quakeml))
            if len(catalog) != 1:
                seisbench.logger.warning(
                    f"Found multiple events in catalog for {event_path.name}. Skipping."
                )
                continue

            event = catalog[0]

            event_params = self._get_event_params(event)

            station_groups = defaultdict(list)
            for pick in event.picks:
                if pick.phase_hint is None:
                    continue

                if pick.waveform_id.network_code == "IA":
                    # Skip restricted IA data
                    continue

                if not isinstance(pick.waveform_id.network_code, str):
                    # Skip traces with invalid network code
                    continue

                station_groups[pick.waveform_id.id[:-1]].append(pick)

            for picks in station_groups.values():
                self._write_picks(
                    picks,
                    event_params,
                    event_path,
                    writer,
                    location_helper,
                    inventory,
                    component_order=component_order,
                )

    @staticmethod
    def _get_event_params(event):
        origin = event.preferred_origin()
        mag = event.preferred_magnitude()
        fm = event.preferred_focal_mechanism()
        event_params = {
            "source_id": str(event.resource_id)[-11:],
            "source_origin_time": str(origin.time),
            "source_origin_uncertainty_sec": origin.time_errors["uncertainty"],
            "source_latitude_deg": origin.latitude,
            "source_latitude_uncertainty_deg": origin.latitude_errors["uncertainty"],
            "source_longitude_deg": origin.longitude,
            "source_longitude_uncertainty_deg": origin.longitude_errors["uncertainty"],
            "source_depth_km": origin.depth / 1e3,
            "source_depth_uncertainty_km": origin.depth_errors["uncertainty"] / 1e3,
        }

        if str(origin.time) < "2012-11-01":
            split = "train"
        elif str(origin.time) < "2013-03-15":
            split = "dev"
        else:
            split = "test"
        event_params["split"] = split

        if mag is not None:
            event_params["source_magnitude"] = mag.mag
            event_params["source_magnitude_uncertainty"] = mag.mag_errors["uncertainty"]
            event_params["source_magnitude_type"] = mag.magnitude_type
            event_params["source_magnitude_author"] = mag.creation_info.agency_id

        if fm is not None:
            try:
                t_axis, p_axis, n_axis = (
                    fm.principal_axes.t_axis,
                    fm.principal_axes.p_axis,
                    fm.principal_axes.n_axis,
                )
                event_params["source_focal_mechanism_t_azimuth"] = t_axis.azimuth
                event_params["source_focal_mechanism_t_plunge"] = t_axis.plunge
                event_params["source_focal_mechanism_t_length"] = t_axis.length

                event_params["source_focal_mechanism_p_azimuth"] = p_axis.azimuth
                event_params["source_focal_mechanism_p_plunge"] = p_axis.plunge
                event_params["source_focal_mechanism_p_length"] = p_axis.length

                event_params["source_focal_mechanism_n_azimuth"] = n_axis.azimuth
                event_params["source_focal_mechanism_n_plunge"] = n_axis.plunge
                event_params["source_focal_mechanism_n_length"] = n_axis.length
            except AttributeError:
                # There seem to be a few broken xml files. In this case, just ignore the focal mechanism.
                pass
        return event_params

    @staticmethod
    def _get_trace_params(picks, location_helper, event_params):
        pick = picks[0]
        lat, lon, elevation, sensitivity = location_helper.find(pick.waveform_id.id)
        if not np.isnan(lat * lon):
            back_azimuth = gps2dist_azimuth(
                event_params["source_latitude_deg"],
                event_params["source_longitude_deg"],
                lat,
                lon,
            )[2]
        else:
            back_azimuth = np.nan
        trace_params = {
            "path_back_azimuth_deg": back_azimuth,
            "station_network_code": pick.waveform_id.network_code,
            "station_code": pick.waveform_id.station_code,
            "trace_channel": pick.waveform_id.channel_code[:-1],
            "station_location_code": pick.waveform_id.location_code,
            "station_latitude_deg": lat,
            "station_longitude_deg": lon,
            "station_elevation_m": elevation,
            "station_sensitivity_counts_spm": sensitivity,
        }
        return trace_params

    def _write_picks(
        self,
        picks,
        event_params,
        event_path,
        writer,
        location_helper,
        inventory,
        time_before=60,
        time_after=60,
        component_order="ZNE",
        suffix="",
    ):
        picks = sorted(picks, key=lambda x: x.time)
        # For stations where spacing between two picks is larger than (time_before + time_after), write multiple traces
        for i, _ in enumerate(picks[:-1]):
            if picks[i + 1].time - picks[i].time > time_before + time_after:
                self._write_picks(
                    picks[: i + 1],
                    event_params,
                    event_path,
                    writer,
                    location_helper,
                    inventory,
                    time_before,
                    time_after,
                    component_order,
                    suffix + "0",
                )
                self._write_picks(
                    picks[i + 1 :],
                    event_params,
                    event_path,
                    writer,
                    location_helper,
                    inventory,
                    time_before,
                    time_after,
                    component_order,
                    suffix + "1",
                )

        trace_params = copy.deepcopy(event_params)
        trace_params.update(
            self._get_trace_params(picks, location_helper, event_params)
        )
        trace_params[
            "trace_name"
        ] = f"{event_params['source_id']}_{picks[0].waveform_id.id[:-1]}{suffix}"

        stream = obspy.Stream()
        loaded = set()
        for pick in picks:
            for c in "ZNE12":
                mseed_path = event_path / f"{pick.waveform_id.id[:-1]}{c}.mseed"
                if str(mseed_path) not in loaded:
                    loaded.add(str(mseed_path))
                    try:
                        stream += obspy.read(str(mseed_path))
                    except FileNotFoundError:
                        pass
                    except InternalMSEEDError:
                        pass
                    except TypeError:
                        # Thrown by obspy for specific invalid MSEED files
                        pass
        stream.merge(-1)

        if len(stream) == 0:
            seisbench.logger.warning(
                f'Found no waveforms for {picks[0].waveform_id.id[:-1]} in event {event_params["source_id"]}'
            )
            return

        sampling_rate = stream[0].stats.sampling_rate
        if any(trace.stats.sampling_rate != sampling_rate for trace in stream):
            seisbench.logger.warning(
                f"Found inconsistent sampling rates for {picks[0].waveform_id.id[:-1]} "
                f'in event {event_params["source_id"]}'
            )
            return

        rotate_stream_to_zne(stream, inventory)

        t_start = min(pick.time for pick in picks) - time_before
        t_end = max(pick.time for pick in picks) + time_after

        stream = stream.slice(t_start, t_end)

        if len(stream) == 0:
            seisbench.logger.warning(
                f'Found no waveforms for {picks[0].waveform_id.id[:-1]} in event {event_params["source_id"]}'
            )
            return

        actual_t_start, data, completeness = stream_to_array(stream, component_order)

        if int((t_end - t_start) * sampling_rate) + 1 > data.shape[1]:
            # Traces appear to be complete, but do not cover the intended time range
            completeness *= data.shape[1] / (int((t_end - t_start) * sampling_rate) + 1)

        trace_params["trace_sampling_rate_hz"] = sampling_rate
        trace_params["trace_completeness"] = completeness
        trace_params["trace_has_spikes"] = trace_has_spikes(data)
        trace_params["trace_start_time"] = str(actual_t_start)
        for pick in picks:
            sample = (pick.time - actual_t_start) * sampling_rate
            trace_params[f"trace_{pick.phase_hint}_arrival_sample"] = sample
            trace_params[f"trace_{pick.phase_hint}_status"] = pick.evaluation_mode

        writer.add_trace(trace_params, data)


class LocationHelper:
    def __init__(self, path):
        self.path = path
        self.full_dict = {}
        self.short_dict = {}
        self.fill_dicts()

    def fill_dicts(self):
        station_list = pd.read_csv(self.path / "station_list.csv")
        for _, row in station_list.iterrows():
            trace_id = row["id"]
            network, station, location, channel = trace_id.split(".")
            if row["sensitivity"] is None or row["sensitivity"] == "None":
                sensitivity = np.nan
            else:
                sensitivity = float(row["sensitivity"])
            self.full_dict[trace_id] = (
                row["lat"],
                row["lon"],
                row["elevation"],
                sensitivity,
            )
            self.short_dict[f"{network}.{station}"] = (
                row["lat"],
                row["lon"],
                row["elevation"],
                np.nan,
            )

        station_list_additional = pd.read_csv(self.path / "station_list_additional.csv")
        for _, row in station_list_additional.iterrows():
            netsta = row["id"]
            self.short_dict[netsta] = (row["lat"], row["lon"], row["elevation"], np.nan)

    def find(self, trace_id):
        if trace_id in self.full_dict:
            return self.full_dict[trace_id]
        else:
            network, station, location, channel = trace_id.split(".")
            netsta = f"{network}.{station}"
            if netsta in self.short_dict:
                return self.short_dict[netsta]
            else:
                return np.nan, np.nan, np.nan, np.nan



import seisbench

import copy
import numpy as np


class FixedWindow:
    """
    A simple windower that returns fixed windows.
    In addition, the windower rewrites all metadata ending in "_sample" to point to the correct sample after window
    selection. Window start and length can be set either at initialization or separately in each call.
    The later is primarily intended for more complicated windowers inheriting from FixedWindow.

    :param p0: Start position of the trace. If p0 is negative, this will be treated as identifying
               a sample before the start of the trace. This is in contrast to standard list indexing
               with negative indices in Python, which counts items from the end of the list. Negative
               p0 is not possible with the strategy "fail".
    :type p0: None or int
    :param windowlen: Window length
    :type windowlen: None or int
    :param strategy: Strategy to mitigate insufficient data. Options are:

                     *  "fail": Raises a ValueError

                     *  "pad": Adds zero padding to the window

                     *  "move": Moves the start to the closest possible position to achieve sufficient trace length.
                        The resulting trace will be aligned to the input trace on one of the ends,
                        depending if parts before (left aligned) or after the trace (right aligned) were requested.
                        Will fail if total trace length is shorter than requested window length.

                     *  "variable": Returns shorter length window, resulting in possibly varying window size.
                        Might return empty window if requested window is completely outside target range.

    :type strategy: str
    :param axis: Axis along which the window selection should be performed
    :type axis: int
    :param key: The keys for reading from and writing to the state dict.
                If key is a single string, the corresponding entry in state dict is modified.
                Otherwise, a 2-tuple is expected, with the first string indicating the key to
                read from and the second one the key to write to.
    :type key: str, tuple[str, str]
    """

    def __init__(self, p0=None, windowlen=None, strategy="fail", axis=-1, key="X"):
        self.p0 = p0
        self.windowlen = windowlen
        self.strategy = strategy
        self.axis = axis
        if isinstance(key, str):
            self.key = (key, key)
        else:
            self.key = key

        if self.strategy not in ["fail", "pad", "move", "variable"]:
            raise ValueError(
                f"Unknown strategy '{self.strategy}'. Options are 'fail', 'pad', 'move', 'variable'"
            )

    def __call__(self, state_dict, p0=None, windowlen=None):
        p0, windowlen = self._validate_parameters(p0, windowlen)

        x, metadata = state_dict[self.key[0]]

        if self.key[0] != self.key[1]:
            # Ensure metadata is not modified inplace unless input and output key are anyhow identical
            metadata = copy.deepcopy(metadata)

        left_padding = None
        p0_padding_offset = 0
        right_padding = None

        if p0 < 0 and self.strategy == "pad":
            left_pad_shape = list(x.shape)
            p0_padding_offset = max(-p0, 0)
            left_pad_shape[self.axis] = min(p0_padding_offset, windowlen)
            left_padding = np.zeros_like(x, shape=left_pad_shape)

            windowlen += p0  # Shorten target window len
            windowlen = max(0, windowlen)  # Windowlen needs to be positive
            p0 = 0

        if x.shape[self.axis] < p0 + windowlen:
            if self.strategy == "fail":
                raise ValueError(
                    f"Requested window length ({windowlen}) is longer than available length after p0 "
                    f"({x.shape[self.axis] - p0})."
                )
            elif self.strategy == "pad":
                p0 = min(p0, x.shape[self.axis])
                old_windowlen = windowlen
                windowlen = x.shape[self.axis] - p0

                right_pad_shape = list(x.shape)
                right_pad_shape[self.axis] = old_windowlen - windowlen
                right_padding = np.zeros_like(x, shape=right_pad_shape)
            elif self.strategy == "move":
                p0 = x.shape[self.axis] - windowlen
                if p0 < 0:
                    raise ValueError(
                        f"Total trace length ({x.shape[self.axis]}) is shorter than requested window length "
                        f"({windowlen})."
                    )
            elif self.strategy == "variable":
                p0 = min(p0, x.shape[self.axis])
                windowlen = x.shape[self.axis] - p0

        window = np.take(x, range(p0, p0 + windowlen), axis=self.axis)

        if left_padding is not None:
            window = np.concatenate([left_padding, window], axis=self.axis)
        if right_padding is not None:
            window = np.concatenate([window, right_padding], axis=self.axis)

        if left_padding is not None:
            p0 -= p0_padding_offset

        for key in metadata.keys():
            if key.endswith("_sample"):
                try:
                    metadata[key] = metadata[key] - p0
                except TypeError:
                    seisbench.logger.info(
                        f"Failed to do window adjustment for column {key} "
                        f"due to type error. "
                    )

        state_dict[self.key[1]] = (window, metadata)

    def _validate_parameters(self, p0, windowlen):
        if p0 is None:
            p0 = self.p0
        if windowlen is None:
            windowlen = self.windowlen

        if p0 is None:
            raise ValueError("Start position must be set in either init or call.")
        if windowlen is None:
            raise ValueError("Window length must be set in either init or call.")

        if p0 < 0 and self.strategy == "fail":
            raise ValueError("Negative indexing is not supported for strategy fail.")

        if p0 < 0 and self.strategy == "move":
            p0 = 0

        if p0 < 0 and self.strategy == "variable":
            windowlen += p0  # Shorten target window len
            windowlen = max(0, windowlen)  # Windowlen needs to be positive
            p0 = 0

        return p0, windowlen

    def __str__(self):
        return f"FixedWindow (p0={self.p0}, windowlen={self.windowlen}, strategy={self.strategy}, axis={self.axis})"


class SlidingWindow(FixedWindow):
    """
    Generates sliding windows and adds a new axis for windows as first axis.
    All metadata entries are converted to arrays of entries.
    Only complete windows are returned and a possible remainder is truncated.
    In particular, if the available data is shorter than the number of windows, an empty array is returned.

    :param timestep: Difference between two consecutive window starts in samples
    :type timestep: int
    :param windowlen: Length of the output window
    :type windowlen: int
    :param kwargs: All kwargs are passed directly to FixedWindow

    """

    def __init__(self, timestep, windowlen, **kwargs):
        self.timestep = timestep
        super().__init__(windowlen=windowlen, **kwargs)

    def __call__(self, state_dict):
        windowlen = self.windowlen
        timestep = self.timestep

        x, metadata = state_dict[self.key[0]]

        if self.key[0] != self.key[1]:
            # Ensure metadata is not modified inplace unless input and output key are anyhow identical
            metadata = copy.deepcopy(metadata)

        n_windows = (x.shape[self.axis] - windowlen) // timestep + 1

        if n_windows == 0:
            target_shape = list(x.shape)
            target_shape[self.axis] = windowlen
            x_out = np.zeros_like(x, shape=[0] + target_shape)
            state_dict[self.key[1]] = (x_out, metadata)
            return

        window_outputs = []
        window_metadatas = []

        for i in range(n_windows):
            tmp_state_dict = {self.key[0]: (x, copy.deepcopy(metadata))}
            super().__call__(tmp_state_dict, p0=i * timestep)
            window_outputs.append(tmp_state_dict[self.key[1]][0])
            window_metadatas.append(tmp_state_dict[self.key[1]][1])

        x_out = np.stack(window_outputs, axis=0)
        collected_metadata = {}
        for key in window_metadatas[0].keys():
            collected_metadata[key] = np.array(
                [window_metadata[key] for window_metadata in window_metadatas]
            )
        state_dict[self.key[1]] = (x_out, collected_metadata)

    def __str__(self):
        return f"SlidingWindow (windowlen={self.windowlen}, timestep={self.timestep})"


class WindowAroundSample(FixedWindow):
    """
    Creates a window around a sample defined in the metadata.
    :param metadata_keys: Metadata key or list of metadata keys to use for window selection.
    The corresponding metadata entries are assumed to be in sample units.
    The generator will return a window starting at the first sample, if all relevant metadata entries are NaN.

    :param samples_before: The number of samples to include before the target sample.
    :type samples_before: int
    :param selection: Selection strategy in case multiple metadata keys are provided and have non-NaN values.
                      Options are:

        - "first": use the first available key
        - "random": use uniform random selection among the keys

    :type selection: str
    :param kwargs: Parameters passed to the init method of FixedWindow.

    """

    def __init__(self, metadata_keys, samples_before=0, selection="first", **kwargs):
        if isinstance(metadata_keys, str):
            self.metadata_keys = [metadata_keys]
        else:
            self.metadata_keys = metadata_keys
        self.samples_before = samples_before
        self.selection = selection

        if selection not in ["first", "random"]:
            raise ValueError(f"Unknown selection strategy '{self.selection}'")

        super().__init__(**kwargs)

    def __call__(self, state_dict, windowlen=None):
        _, metadata = state_dict[self.key[0]]
        cand = [
            metadata[key]
            for key in self.metadata_keys
            if key in metadata and not np.isnan(metadata[key])
        ]

        if len(cand) == 0:
            cand = [self.samples_before]
            # raise ValueError("Found no possible candidate for window selection")

        if self.selection == "first":
            p0 = cand[0]
        elif self.selection == "random":
            p0 = np.random.choice(cand)
        else:
            raise ValueError(f"Unknown selection strategy '{self.selection}'")

        p0 -= self.samples_before

        super(WindowAroundSample, self).__call__(
            state_dict, p0=int(p0), windowlen=windowlen
        )

    def __str__(self):
        return (
            f"WindowAroundSample (metadata_keys={self.metadata_keys}, samples_before={self.samples_before}, "
            f"selection={self.selection})"
        )


class RandomWindow(FixedWindow):
    """
    Selects a window within the range [low, high) randomly with a uniform distribution.
    If there are no candidates fulfilling the criteria, the window selection depends on the strategy.
    For "fail" or "move" a ValueError is raise.
    For "variable" only the part between low and high is returned. If high < low, this part will be empty.
    For "pad" the same as for "variable" will be returned, but padded to the correct length.
    The padding will be added randomly to both sides.

    :param low: The lowest allowed index for the start sample.
                The sample at this position can be included in the output.
    :type low: None or int
    :param high: The highest allowed index for the end.
                 The sample at position high can *not* be included in the output
    :type high: None or int
    :param kwargs: Parameters passed to the init method of FixedWindow.
    """

    def __init__(self, low=None, high=None, windowlen=None, **kwargs):
        super().__init__(windowlen=windowlen, **kwargs)

        self.low = low
        self.high = high

        if high is not None:
            if low is None:
                low = 0  # Temporary value for validity check that is not stored

            if windowlen is not None:
                if high - low < windowlen and self.strategy in ["fail", "move"]:
                    raise ValueError(
                        "Difference between low and high must be at least window length."
                    )
            elif low >= high:
                raise ValueError("Low value needs to be smaller than high value.")

    def __call__(self, state_dict, windowlen=None):
        x, _ = state_dict[self.key[0]]
        _, windowlen = self._validate_parameters(0, windowlen)

        low, high = self.low, self.high
        if high is None:
            high = x.shape[self.axis]
        if low is None:
            low = 0

        if low > high - windowlen:
            if self.strategy == "pad":
                # Get larger interval around target window with random location
                # The parent FixedWindow will handle possible cases in which the waveform is too short
                p0 = np.random.randint(high - windowlen, low + 1)
                super().__call__(state_dict, p0=p0, windowlen=windowlen)
                # Get output and overwrite padded range with zeros
                x, metadata = state_dict[self.key[1]]

                def fill_range_with_zeros(a, b):
                    # Fills range of array with zeros
                    # Automatically selects the correct axis
                    axis = self.axis
                    if axis < 0:
                        axis = x.ndim + axis

                    ind = np.arange(a, b, dtype=int)
                    for _ in range(axis):
                        ind = np.expand_dims(ind, 0)
                    for _ in range(axis + 1, x.ndim):
                        ind = np.expand_dims(ind, -1)

                    np.put_along_axis(x, ind, 0, axis)

                fill_range_with_zeros(0, low - p0)
                fill_range_with_zeros(high - p0, x.shape[self.axis])
                return

            elif self.strategy == "variable":
                # Return interval between high and low
                p0 = low
                windowlen = max(0, high - low)
                super().__call__(state_dict, p0=p0, windowlen=windowlen)
                return

            else:
                raise ValueError("No possible candidates for random window selection.")

        p0 = np.random.randint(low, high - windowlen + 1)
        super().__call__(state_dict, p0=p0, windowlen=windowlen)

    def __str__(self):
        return f"RandomWindow (low={self.low}, high={self.high})"


class SteeredWindow(FixedWindow):
    """
    Window selection that relies on the "_control_" dict from the :py:class:`SteeredGenerator`.
    Selects a window of given length with the window defined by start and end sample in the middle.
    If no length is given, the window between start and end sample is returned.
    If there are insufficient samples on one side of the target window, the window will be moved.
    If the total number of samples is insufficient, the window will start at the earliest possible sample.
    The behavior in this case depends on the chosen strategy for :py:class:`FixedWindow`.

    :param windowlen: Length of the window to be returned.
                      If None, will be determined using the start and end samples from the "_control_" dict.
    :type windowlen: int or None
    :param start_key: Key of the start sample in the "_control_" dict
    :type start_key: str
    :param end_key: Key of the end sample in the "_control_" dict
    :type end_key: str
    :param window_output_key: The sample start and end will be written as numpy array to this key in the state_dict
    :type window_output_key: str
    :param kwargs: Parameters passed to the init method of FixedWindow.
    """

    def __init__(
        self,
        windowlen,
        start_key="start_sample",
        end_key="end_sample",
        window_output_key="window_borders",
        **kwargs,
    ):
        super().__init__(windowlen=windowlen, **kwargs)

        self.start_key = start_key
        self.end_key = end_key
        self.window_output_key = window_output_key

    def __call__(self, state_dict):
        control = state_dict["_control_"]

        start_sample = int(control[self.start_key])
        end_sample = int(control[self.end_key])

        if self.windowlen is None:
            windowlen = end_sample - start_sample
        else:
            windowlen = self.windowlen

        x, _ = state_dict[self.key[0]]
        n_samples = x.shape[self.axis]

        sample_range = end_sample - start_sample
        if sample_range > windowlen:
            raise ValueError(
                f"Requested window length is {windowlen}, but the difference between start and end sample "
                f"is already {end_sample - start_sample}. Can't return window."
            )

        elif sample_range == windowlen:
            p0 = max(0, start_sample)

        else:
            p0 = (
                start_sample - (windowlen - sample_range) // 2
            )  # Try to put the window into the center

            if p0 + windowlen > n_samples:
                p0 = (
                    n_samples - windowlen
                )  # If the window end is behind the trace end, move it backward

            p0 = max(
                0, p0
            )  # If the window start is before the trace start, move it forward

        window_borders = np.array([start_sample - p0, end_sample - p0])
        state_dict[self.window_output_key] = (
            window_borders,
            {},
        )  # No metadata associated

        super().__call__(state_dict, p0=p0, windowlen=windowlen)

    def __str__(self):
        return f"SteeredWindow"



import seisbench

import requests
import ftplib
from tqdm import tqdm
from pathlib import Path
import time
import os
from lxml import etree


def download_http(
    url, target, progress_bar=True, desc="Downloading", precheck_timeout=3
):
    """
    Downloads file from http/https source. Raises a ValueError for non-200 status codes.

    :param url: Target url
    :type url: str
    :param target: Path to save to
    :type target: Path or str
    :param progress_bar: If true, shows a progress bar for the download
    :type progress_bar: bool
    :param desc: Description for the progress bar
    :type desc: str
    :param precheck_timeout: Timeout passed to :py:func:`precheck_url`
    :type precheck_timeout: int
    """
    seisbench.logger.info(f"Downloading file from {url} to {target}")

    precheck_url(url, timeout=precheck_timeout)

    req = requests.get(url, stream=True, headers={"User-Agent": "SeisBench"})

    if req.status_code != 200:
        raise ValueError(
            f"Invalid URL. Request returned status code {req.status_code}."
        )

    content_length = req.headers.get("Content-Length")
    total = int(content_length) if content_length is not None else None
    if progress_bar:
        pbar = tqdm(
            unit="B", total=total, desc=desc, unit_scale=True, unit_divisor=1024
        )
    else:
        pbar = None

    target = Path(target)

    with open(target, "wb") as f_target:
        for chunk in req.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                if pbar is not None:
                    pbar.update(len(chunk))
                f_target.write(chunk)

    if progress_bar:
        pbar.close()


def precheck_url(url, timeout):
    """
    Checks whether the url is reachable and give a 200 or 300 HTTP response code.
    If a timeout occurs or a >=400 response code is returned, the precheck issues a warning.

    :param url: URL to check
    :param timeout: Timeout in seconds
    """
    if timeout <= 0:
        # Skip check
        return

    error_description = None

    try:
        req = requests.head(url, timeout=timeout, headers={"User-Agent": "SeisBench"})

        if req.status_code >= 400:  # Assumes all 200 and 300 codes are acceptable
            error_description = f"status code {req.status_code}"

    except requests.Timeout:
        error_description = "a timeout"

    except requests.ConnectionError:
        error_description = "a connection error"

    if error_description is not None:
        seisbench.logger.warning(
            f"The download precheck failed with {error_description}. "
            f"This is not an error itself, but might indicate a subsequent error. "
            f"If you encounter an error, this might be caused by the firewall setup of your "
            f"network. "
            f"Please check https://github.com/seisbench/seisbench#known-issues for details."
        )


def download_ftp(
    host,
    file,
    target,
    user="anonymous",
    passwd="",
    blocksize=8192,
    progress_bar=True,
    desc="Downloading",
):
    """
    Downloads file from ftp source.

    :param host: Host URL
    :type host: str
    :param file: File path on the FTP server
    :type file: str
    :param target: Path to save to
    :type target: Path or str
    :param user: Username for login
    :type user: str
    :param passwd: Password for login
    :type passwd: str
    :param blocksize: Size of download blocks in bytes
    :type blocksize: int
    :param progress_bar: If true, shows a progress bar for the download
    :type progress_bar: bool
    :param desc: Description for the progress bar
    :type desc: str
    """
    with ftplib.FTP(host, user, passwd) as ftp:
        ftp.voidcmd("TYPE I")
        total = ftp.size(file)

        if progress_bar:
            pbar = tqdm(
                unit="B", total=total, desc=desc, unit_scale=True, unit_divisor=1024
            )

        def callback(chunk):
            if progress_bar:
                pbar.update(len(chunk))
            fout.write(chunk)

        with open(target, "wb") as fout:
            ftp.retrbinary(f"RETR {file}", callback, blocksize=blocksize)

        if progress_bar:
            pbar.close()


def ls_webdav(url, precheck_timeout=3):
    """
    Lists the files in a WebDAV directory

    :param url: URL of the directory to list
    :type url: str
    :param precheck_timeout: Timeout passed to :py:func:`precheck_url`
    :type precheck_timeout: int
    :return: List of files
    """
    precheck_url(url, timeout=precheck_timeout)

    xml_request = b'<?xml version="1.0"?><a:propfind xmlns:a="DAV:"><a:prop><a:resourcetype/></a:prop></a:propfind>'

    req = requests.Request("PROPFIND", url, headers={"Depth": "1"}, data=xml_request)
    prep = req.prepare()

    with requests.Session() as sess:
        ret = sess.send(prep)

    if not 200 <= ret.status_code < 300:
        raise ValueError(
            f"Invalid URL. Request returned status code {ret.status_code}."
        )

    files = ["."]  # The first part of the response is always the root path
    tree = etree.fromstring(ret.content)
    for elem in tree.xpath(".//d:href", namespaces={"d": "DAV:"})[1:]:
        # Extract basepath from path
        p = elem.text[:-1].rfind("/") + 1
        files.append(elem.text[p:])

    return files


def callback_if_uncached(
    files, callback, force=False, wait_for_file=False, test_interval=60
):
    """
    Checks if all files exists and executes the callback otherwise.
    Please note that the callback is executed if *at least one* file is not cached.
    If one of the files does not exists, but file.partial does, the behaviour depends on force and wait_for_file.

    .. warning::
        While making concurrent callbacks unlikely, they can still happen, if the function is called twice in short
        time, i.e., the second starts before the first created a .partial file.

    :param files: A list of files or single file to check.
    :type files: list[union[Path, str]], Path, str
    :param callback: A callback, taking one parameter, a list of target file names. Will be called if a file is missing.
                     The callback will be given the same parameter as provided in files, just with files renamed
                     to file.partial. The function will move the files afterwards, but will ignore empty files.
    :type callback: callable
    :param force: If true, and not all files exist, ignore and remove all partial files and execute callback. Only use
                  this parameter if no other instance of callback_if_uncached is currently requesting the same file.
    :type force: bool
    :param wait_for_file: If true, not all files exist, but partial files exist, sleep until files exists or no partial
                          files exist.
    :type wait_for_file: bool
    :param test_interval: Sleep interval for wait_for_file.
    :type test_interval: float
    """
    if not isinstance(files, (list, tuple)):
        files = [files]
        squeeze = True
    else:
        squeeze = False

    files = [Path(file) for file in files]
    partial_files = [file.parent / (file.name + ".partial") for file in files]

    def exist():
        return all(file.is_file() for file in files)

    def is_partial():
        return any(file.is_file() for file in partial_files)

    while not exist() and is_partial() and not force:
        if wait_for_file:
            seisbench.logger.warning(
                f"Found partial instance. Rechecking in {test_interval} seconds."
            )
            time.sleep(test_interval)
        else:
            raise ValueError(
                f"Found partial instance. "
                f"This suggests that either the download is currently in progress or a download failed. "
                f"To redownload the file, call the dataset with force=True. "
                f"To wait for another download to finish, use wait_for_file=True."
            )

    if not exist() and force:
        for file in files:
            partial_file = file.parent / (file.name + ".partial")
            if partial_file.is_file():
                os.remove(partial_file)

    if exist():
        return
    else:
        # Open and close each partial file once, to ensure they exist early and make race conditions unlikely.
        for file in partial_files:
            file.parent.mkdir(parents=True, exist_ok=True)
            open(file, "a").close()

        try:
            if squeeze:
                callback(partial_files[0])
            else:
                callback(partial_files)
        finally:
            for partial_file in partial_files:
                if partial_file.stat().st_size == 0:
                    os.remove(partial_file)

        seisbench.logger.info("Moving partial files to target")
        for partial_file, file in zip(partial_files, files):
            if partial_file.is_file():
                partial_file.rename(file)



import seisbench

import numpy as np
from obspy import ObsPyException


def trace_has_spikes(data, factor=25, quantile=0.975):
    """
    Checks for bit flip errors in the data using a simple quantile rule

    :param data: Data array
    :type data: np.ndarray
    :param factor: Maximum allowed factor between peak and quantile
    :type factor: float
    :param quantile: Quantile to check. Must be between 0 and 1.
    :type quantile: float
    """
    q = np.quantile(np.abs(data), quantile, axis=1, keepdims=True)
    return np.any(data > q * factor)


def stream_to_array(stream, component_order):
    """
    Converts stream of single station waveforms into a numpy array according to a given component order.
    If trace start and end times disagree between component traces, remaining parts are filled with zeros.
    Also returns completeness, i.e., the fraction of samples in the output that actually contain data.
    Assumes all traces to have the same sampling rate.

    :param stream: Stream to convert
    :type stream: obspy.Stream
    :param component_order: Component order
    :type component_order: str
    :return: starttime, data, completeness
    :rtype: UTCDateTime, np.ndarray, float
    """
    starttime = min(trace.stats.starttime for trace in stream)
    endtime = max(trace.stats.endtime for trace in stream)
    sampling_rate = stream[0].stats.sampling_rate

    samples = int((endtime - starttime) * sampling_rate) + 1

    completeness = 0.0
    data = np.zeros((len(component_order), samples), dtype="float64")
    for c_idx, c in enumerate(component_order):
        c_stream = stream.select(channel=f"??{c}")
        if len(c_stream) > 1:
            # If multiple traces are found, issue a warning and write them into the data ordered by their length
            seisbench.logger.warning(
                f"Found multiple traces for {c_stream[0].id} starting at {stream[0].stats.starttime}. "
                f"Completeness will be wrong in case of overlapping traces."
            )
            c_stream = sorted(c_stream, key=lambda x: x.stats.npts)

        c_completeness = 0.0
        for trace in c_stream:
            start_sample = int((trace.stats.starttime - starttime) * sampling_rate)
            l = min(len(trace.data), samples - start_sample)
            data[c_idx, start_sample : start_sample + l] = trace.data[:l]
            c_completeness += l

        completeness += min(1.0, c_completeness / samples)

    data -= np.mean(data, axis=1, keepdims=True)

    completeness /= len(component_order)
    return starttime, data, completeness


def rotate_stream_to_zne(stream, inventory):
    """
    Tries to rotate the stream to ZNE inplace. There are several possible failures, which are silently ignored.

    :param stream: Stream to rotate
    :type stream: obspy.Stream
    :param inventory: Inventory object
    :type inventory: obspy.Inventory
    """
    try:
        stream.rotate("->ZNE", inventory=inventory)
    except ValueError:
        pass
    except NotImplementedError:
        pass
    except AttributeError:
        pass
    except ObsPyException:
        pass
    except Exception:
        # Required, because obspy throws a plain Exception for missing channel metadata
        pass


def waveform_id_to_network_station_location(waveform_id):
    """
    Takes a waveform_id as string in the format Network.Station.Location.Channel and
    returns a string with channel dropped. If the waveform_id does not conform to the format,
    the input string is returned.

    :param waveform_id: Waveform ID in format Network.Station.Location.Channel
    :type waveform_id: str
    :return: Waveform ID in format Network.Station.Location
    :rtype: str
    """
    parts = waveform_id.split(".")
    if len(parts) != 4:
        return waveform_id
    else:
        return ".".join(parts[:-1])



import seisbench
import seisbench.util
from .base import BenchmarkDataset

import pandas as pd
import numpy as np
from obspy import UTCDateTime
from obspy.clients.fdsn import Client


class DummyDataset(BenchmarkDataset):
    """
    A dummy dataset visualizing the implementation of custom datasets
    """

    def __init__(self, **kwargs):
        citation = (
            "Münchmeyer, Jannes; Bindi, Dino; Sippl, Christian; Leser, Ulf; Tilmann, Frederik (2019): "
            "Magnitude scales, attenuation models and feature matrices for the IPOC catalog. "
            "V. 1.0. GFZ Data Services. https://doi.org/10.5880/GFZ.2.4.2019.004"
        )
        super().__init__(citation=citation, repository_lookup=True, **kwargs)

    def _download_dataset(self, writer, trace_length=60, **kwargs):
        sampling_rate = 20

        writer.metadata_dict = {
            "time": "trace_start_time",
            "latitude": "source_latitude_deg",
            "longitude": "source_longitude_deg",
            "depth": "source_depth_km",
            "cls": "source_event_category",
            "MA": "source_magnitude",
            "ML": "source_magnitude2",
            "std_MA": "source_magnitude_uncertainty",
            "std_ML": "source_magnitude_uncertainty2",
        }
        writer.data_format = {
            "dimension_order": "CW",
            "component_order": "ZNE",
            "sampling_rate": sampling_rate,
            "measurement": "velocity",
            "unit": "counts",
            "instrument_response": "not restituted",
        }

        path = self.path
        path.mkdir(parents=True, exist_ok=True)

        seisbench.util.download_ftp(
            "datapub.gfz-potsdam.de",
            "download/10.5880.GFZ.2.4.2019.004/IPOC_catalog_magnitudes.csv",
            path / "raw_catalog.csv",
            progress_bar=False,
        )

        metadata = pd.read_csv(path / "raw_catalog.csv")
        metadata = metadata.iloc[:100].copy()

        def to_tracename(x):
            for c in "/:.":
                x = x.replace(c, "_")
            return x

        client = Client("GFZ")
        inv = client.get_stations(
            net="CX",
            sta="PB01",
            starttime=UTCDateTime.strptime(
                "2007/01/01 00:00:00.00", "%Y/%m/%d %H:%M:%S.%f"
            ),
        )

        metadata["trace_name"] = metadata["time"].apply(to_tracename)
        metadata["station_network_code"] = "CX"
        metadata["station_code"] = "PB01"
        metadata["station_type"] = "BH"
        metadata["station_latitude_deg"] = inv[0][0].latitude
        metadata["station_longitude_deg"] = inv[0][0].longitude
        metadata["station_elevation_m"] = inv[0][0].elevation
        metadata["source_magnitude_type"] = "MA"
        metadata["source_magnitude_type2"] = "ML"

        splits = 60 * ["train"] + 10 * ["dev"] + 30 * ["test"]
        metadata["split"] = splits

        writer.set_total(len(metadata))
        for _, row in metadata.iterrows():
            time = row["time"]
            waveform = np.zeros((3, sampling_rate * trace_length))
            time = UTCDateTime.strptime(time, "%Y/%m/%d %H:%M:%S.%f")
            stream = client.get_waveforms(
                "CX", "PB01", "*", "BH?", time, time + trace_length
            )
            for cid, component in enumerate("ZNE"):
                ctrace = stream.select(channel=f"??{component}")[0]
                waveform[cid] = ctrace.data[: sampling_rate * trace_length].astype(
                    float
                )
            writer.add_trace(row, waveform)


class ChunkedDummyDataset(BenchmarkDataset):
    """
    A chunked dummy dataset visualizing the implementation of custom datasets with chunking
    """

    def __init__(self, **kwargs):
        citation = (
            "Münchmeyer, Jannes; Bindi, Dino; Sippl, Christian; Leser, Ulf; Tilmann, Frederik (2019): "
            "Magnitude scales, attenuation models and feature matrices for the IPOC catalog. "
            "V. 1.0. GFZ Data Services. https://doi.org/10.5880/GFZ.2.4.2019.004"
        )

        # Write chunks to file
        chunks_path = self.path / "chunks"
        if not chunks_path.is_file():
            self.path.mkdir(exist_ok=True, parents=True)
            with open(chunks_path, "w") as f:
                f.write("0\n1\n")

        super().__init__(citation=citation, repository_lookup=True, **kwargs)

    def _download_dataset(self, writer, chunk, trace_length=60, **kwargs):
        sampling_rate = 20

        writer.metadata_dict = {
            "time": "trace_start_time",
            "latitude": "source_latitude_deg",
            "longitude": "source_longitude_deg",
            "depth": "source_depth_km",
            "cls": "source_event_category",
            "MA": "source_magnitude",
            "ML": "source_magnitude2",
            "std_MA": "source_magnitude_uncertainty",
            "std_ML": "source_magnitude_uncertainty2",
        }
        writer.data_format = {
            "dimension_order": "CW",
            "component_order": "ZNE",
            "sampling_rate": sampling_rate,
            "measurement": "velocity",
            "unit": "counts",
            "instrument_response": "not restituted",
        }

        path = self.path
        path.mkdir(parents=True, exist_ok=True)

        seisbench.util.download_ftp(
            "datapub.gfz-potsdam.de",
            "download/10.5880.GFZ.2.4.2019.004/IPOC_catalog_magnitudes.csv",
            path / "raw_catalog.csv",
            progress_bar=False,
        )

        metadata = pd.read_csv(path / "raw_catalog.csv")
        if chunk == "0":
            metadata = metadata.iloc[:100].copy()
        elif chunk == "1":
            metadata = metadata.iloc[100:200].copy()
        else:
            raise ValueError(f'Unknown chunk "{chunk}"')

        def to_tracename(x):
            for c in "/:.":
                x = x.replace(c, "_")
            return x

        client = Client("GFZ")
        inv = client.get_stations(
            net="CX",
            sta="PB01",
            starttime=UTCDateTime.strptime(
                "2007/01/01 00:00:00.00", "%Y/%m/%d %H:%M:%S.%f"
            ),
        )

        metadata["trace_name"] = metadata["time"].apply(to_tracename)
        metadata["station_network_code"] = "CX"
        metadata["station_code"] = "PB01"
        metadata["station_type"] = "BH"
        metadata["station_latitude_deg"] = inv[0][0].latitude
        metadata["station_longitude_deg"] = inv[0][0].longitude
        metadata["station_elevation_m"] = inv[0][0].elevation
        metadata["source_magnitude_type"] = "MA"
        metadata["source_magnitude_type2"] = "ML"

        splits = 60 * ["train"] + 10 * ["dev"] + 30 * ["test"]
        metadata["split"] = splits

        writer.set_total(len(metadata))
        for _, row in metadata.iterrows():
            time = row["time"]
            waveform = np.zeros((3, sampling_rate * trace_length))
            time = UTCDateTime.strptime(time, "%Y/%m/%d %H:%M:%S.%f")
            stream = client.get_waveforms(
                "CX", "PB01", "*", "BH?", time, time + trace_length
            )
            for cid, component in enumerate("ZNE"):
                ctrace = stream.select(channel=f"??{component}")[0]
                waveform[cid] = ctrace.data[: sampling_rate * trace_length].astype(
                    float
                )
            writer.add_trace(row, waveform)



import seisbench
from .base import BenchmarkDataset


class Iquique(BenchmarkDataset):
    """
    Iquique Benchmark Dataset of local events used for training in Woollam (2019)
    study (see citation).

    Splits are set using standard random sampling of :py:class:`seisbench.data.base.BenchmarkDataset`.
    """

    def __init__(self, **kwargs):

        citation = (
            "Woollam, J., Rietbrock, A., Bueno, A. and De Angelis, S., 2019. "
            "Convolutional neural network for seismic phase classification, "
            "performance demonstration over a local seismic network. "
            "Seismological Research Letters, 90(2A), pp.491-502. "
            "https://doi.org/10.1785/0220180312"
        )

        seisbench.logger.warning(
            "Check available storage and memory before downloading and general use "
            "of Iquique dataset. "
            "Dataset size: waveforms.hdf5 ~5Gb, metadata.csv ~2.6Mb"
        )

        super().__init__(citation=citation, repository_lookup=True, **kwargs)

    def _download_dataset(**kwargs):
        pass



import os
from pathlib import Path
from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


with open(Path(os.path.dirname(__file__)) / "requirements.txt") as f:
    required = f.readlines()

setup(
    name="seisbench",
    version="0.2.1",
    author="Jack Woollam,Jannes Münchmeyer",
    author_email="jack.woollam@kit.edu,munchmej@gfz-potsdam.de",
    description="The seismological machine learning benchmark collection",
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seisbench/seisbench",
    packages=find_packages(exclude="tests"),
    python_requires=">=3.7",
    install_requires=required,
)
